package Product;

$Product::Version = '0.1.0';

require Exporter;
@ISA = qw( Exporter );
@EXPORT= qw( build_released_files_list
	     build_licence_excluded_files_list
	     build_excluded_from_tag_checks_files_list
	     build_files_list
	     remove_from_files_list
	     check_arguments
	     parse_arguments_no_such_option
	     file_content
	     convert_to_valid_regular_expression
	     in_tag_exclude_list
	     contains_licence
	     change_licence );

use strict;
use Data::Dumper;

# Exit codes
my $EXIT_USAGE=-1;
my $EXIT_CONFIGURATION_FILE_DOES_NOT_EXIST=-5;

#
# Builds a list of all files that will be released.
#
# PARAMETERS:
# $1 - Reference to a hash containing key-value pairs describing the
#      configuration of a tool that will use this function
#
# RETURNS: A reference to a hash containing the list of all files that
#          will be released
sub build_released_files_list {
    my ( $configuration ) = ( @_ );
    my $files_list_included_regex;
    my $files_list_included;
    my $files_list_excluded_regex;
    my $files_list_excluded;
    my $files_list_final;

    # Build a list of regular expressions used to describe files that
    # will be included in the release
    foreach my $file_regex ( @{ $configuration->{ DEVEL_PRODUCT_RELEASE_CONFIG_RELEASE_FILES_LIST_INCLUDED } } ) {
	push @$files_list_included_regex, $file_regex;
    }

    # Build a complete list of files out of the included files
    $files_list_included = build_files_list( $files_list_included_regex );

    # Build a list of regular expressions used to describe files that
    # will be excluded from the release
    foreach my $file_regex ( @{ $configuration->{ DEVEL_PRODUCT_RELEASE_CONFIG_RELEASE_FILES_LIST_EXCLUDED } } ) {
	push @$files_list_excluded_regex, $file_regex;
    }

    # Build a complete list of files out of the excluded files
    $files_list_excluded = build_files_list( $files_list_excluded_regex );

    # Remove from the list of included files, the list of excluded files
    $files_list_final = $files_list_included;
    foreach my $excluded_file ( sort keys %$files_list_excluded ) {
	delete $files_list_final->{ $excluded_file };
    }

    # print "Files included are " . Dumper( $files_list_included );
    # print "Files excluded are " . Dumper( $files_list_excluded );
    # print "Files final are " . Dumper( $files_list_final );

    return $files_list_final;
}



#
# Builds a list of all files that will be excluded from checks for a valid licence
#
# $1 - Reference to a hash containing key-value pairs describing the
#      configuration of a tool that will use this function
#
# RETURNS: A reference to a hash containing the list of all files that
#          will be excluded from checks for a valid licence
sub build_licence_excluded_files_list {
    my ( $configuration ) = ( @_ );
    my $files_list_excluded_regex;
    my $files_list_excluded;
    my $files_list_final;


    # Build a list of regular expressions used to describe files that
    # will be excluded from checks for a valid licence
    foreach my $file_regex ( @{ $configuration->{ DEVEL_PRODUCT_RELEASE_CONFIG_COPYRIGHT_LICENCE_FILES_LIST_EXCLUDED } } ) {
	push @$files_list_excluded_regex, $file_regex;
    }

    # Build a complete list of files out of the excluded files
    $files_list_excluded = Product::build_files_list( $files_list_excluded_regex );

    # print Dumper( $files_list_excluded );

    return $files_list_excluded;
}



#
# Builds a list of all files that will be excluded from tag checks
#
# PARAMETERS:
# $1 - Reference to a hash containing key-value pairs describing the
#      configuration of a tool that will use this function
#
# RETURNS: A reference to a hash containing the list of all files that
#          will be excluded from tag checks released
sub build_excluded_from_tag_checks_files_list {
    my ( $configuration ) = ( @_ );
    my $files_list_excluded_regex;
    my $files_list_excluded;

    # Build a list of regular expressions used to describe files that
    # will be excluded from tag checks in the release
    foreach my $file_regex ( @{ $configuration->{ DEVEL_PRODUCT_RELEASE_CONFIG_RELEASE_TAG_FILES_LIST_EXCLUDED } } ) {
	push @$files_list_excluded_regex, $file_regex;
    }

    # Build a complete list of files that will be excluded from tag
    # checks out of the excluded files regular expressions
    $files_list_excluded = build_files_list( $files_list_excluded_regex );

    return $files_list_excluded;
}



# Builds a list of files based on an array of regular expressions
# describing that list.
#
# PARAMETERS:
# $1 - Reference to an array containing the list of file descriptions
#      as regular expressions
#
# RETURNS:
# A reference to a hash containing the list of files matching the
# regular expressions array

sub build_files_list {
    my $files_regex_list = shift;
    my $files_list;

    # Grab the listing of all files. Exclude directories
    open( FILES_LIST, 'find . ! \( -type d \) -print|');

    # Check which files will be added in the final list of files
    while( my $line = <FILES_LIST>) {
	# print $line;
	$line =~ s/\.\///g; # Strip starting "./"
	$line =~ s/\n/ /;   # Strip trailing newline
	$line =~ s/\s*$//;  # Strip trailing whitespace

	foreach my $regex ( @$files_regex_list ) {
	    if ( $line =~ m/$regex/ ) {
		$files_list->{ $line } = 1;
	    }
	}
    }

    # Return the list of files build up
    return $files_list;
}



#
# Given two hashes of files, removes from the first hash the files of
# the second hash.
#
# PARAMETERS:
# $1 - hash reference of the first file list
# $2 - hash reference of the second file list
#
# RETURNS:
# A reference to a hash containing the files in the first hash that
# are not in the second hash

sub remove_from_files_list {
    my ( $first, $second ) = ( @_ );
    my $files_list_final;

    # Remove from the list of files, the list of excluded files
    $files_list_final = $first;
    foreach my $excluded_file ( sort keys %$second ) {
	delete $files_list_final->{ $excluded_file };
    }

    return $files_list_final;
}



#
# Checks if the arguments supplied are valid
#
# PARAMETERS:
# $1 - hash reference in which all command-line arguments are stored 
sub check_arguments {
    my ( $arguments ) = ( @_ );
    my $conf;

    # If no arguments where supplied, display the usage information
    if ( $arguments->{ "config_file" } eq "" ) {
	usage();
    }

    # load the configuration for running this tool
    $conf = load_configuration( $arguments->{ "config_file" } );

    return $conf;
}



#
# Displays information on how to use this tool
#
sub usage {    
    print "
Usage: $0 [arguments]
Valid arguments are:
  --config=<filename>		name of configuration file to process
  --help			this help message
  --version			print version
";
	
    exit $EXIT_USAGE;
}



#
# Loads the configuration file for running this tool
#
# PARAMETERS:
# $1 - name of configuration file
sub load_configuration {
    my ( $filename ) = ( @_ );
    my $content = "";
    my $conf;

    # Read the configuration file content
    if ( ! -e "$filename" ) {
	die "Configuration file \"$filename\" does not exist.\n";
	exit $EXIT_CONFIGURATION_FILE_DOES_NOT_EXIST;
    }

    open  CONFIGURATION_FILE, "$filename";
    while ( my $line = <CONFIGURATION_FILE>) {
	$content .= $line;
    }
    close CONFIGURATION_FILE;

    # Massage the content so it becames a valid Perl hash
    $content = "\$conf = {
$content
\"DUMMY_SYMBOL\" => \"DUMMY_VALUE\" };
";

    # Evaluate the content as valid Perl source. This should load the configuration file's values
    eval $content;

    # Remove the dummy symbol
    delete %$conf->{ DUMMY_SYMBOL };

    return $conf;
}



#
# Handles the case where an invalid command-line argument is parsed
#
sub parse_arguments_no_such_option {
    my ( $argument_name ) = ( @_ );

    print "Command-line argument $argument_name is invalid\n";
    usage();
}



#
# Returns the contents of a file
#
# PARAMETERS:
# $1 - Name of file whose contents will be read
#
# RETURNS:
# Contents of file
sub file_content {
    my ( $filename ) = ( @_ );
    my $contents;

    {
	$/ = undef;               # Slurp in whole file
	open ( FILE_CONTENT, "$filename" ) || die "File \"$filename\" cannot be opened\n";
	$contents = <FILE_CONTENT>;
	close ( FILE_CONTENT );
    }

    return $contents;
}



#
# Converts a string to a valid regular expression
#
# PARAMETERS:
# $1 - The string that will be converted
#
# RETURNS:
# The converted string
sub convert_to_valid_regular_expression {
    my ( $content ) = ( @_ );
    my $converted_content;

    $converted_content = $content;

    # Escape all these symbols
    my $escapes = { "@" => 1,
		    "(" => 1,
		    ")" => 1,
		    "/" => 1,
		    "." => 1,
		    "-" => 1,
		};

    # Convert all them to their escaped versions
    foreach my $symbol ( sort keys %$escapes ) {
	my $from = "\\" . $symbol;
	my $to = "\\" . $symbol;

	$converted_content =~ s/$from/$to/gs;
    }

#    print "\nCONTENT WAS: $content\n";
#    print "\nCONVERTED CONTENT WAS: $converted_content\n";
#    exit 0;

    return $converted_content;
}




#
# Checks if file is in the list of files that should be excluded from
# tag checks
#
# PARAMETERS:
# $1 - The file to check
# $2 - A hash that holds the list of files that should be excluded
#      from tag checks.
#
# RETURNS: 1 if the file is in the list,
#          0 otherwise
sub in_tag_exclude_list {
    my ( $file, $exclude_list ) = ( @_ );

    return in_hash( $file, $exclude_list );
}


# Checks if an element is a key in a hash
#
# $1 - The element that will be checked
# $2 - The hash
#
# RETURNS: 1 if the element is in the hash,
#          0 otherwise
sub in_hash {
    my ( $element, $hash ) = ( @_ );
    my $retval = 0;

    foreach my $hash_element ( sort keys %$hash ) {
#	print "hash element is '$hash_element' and element is '$element'\n";
	if ( $hash_element eq $element ) {
	    $retval = 1;
	    goto out;
	}
    }

out:
    return $retval;

}



# Checks if a file contains a licence
#
# $1 - The file content
# $2 - The licence, already converted to a valid regular expression
#
# RETURNS: 2 arguments:
# First:   1 if the licence is missing
#          2 if the licence exists but is invalid
#          0 if the licence is valid
# Second:  The content that followed the licence
sub contains_licence {
    my ( $file_content, $licence_content )= ( @_ );
    my $invalid_licence = 0;
    my $prefix_content = "";
    my $postfix_content = "";

    # Check if the file contains the licence. Need to account for the
    # licence being present in files that are Makefiles (escaped by
    # #s), C sources, (escaped by /*s or *s), etc. As a result we are
    # looking to find a match for the first line of the licence in a
    # file, and then inspect if each subsequent line in the licence
    # matches subsequent lines in the file

    # Extract the first line of the licence
    my $licence_content_first_line = $licence_content;
    if ( $licence_content_first_line =~ m/^(.*)$/m ) {
	$licence_content_first_line = $1;
    }
    
    #print "\nfirst line is " . $licence_content_first_line;
    
    # Extract the individual lines of the file
    my ( @file_lines ) = split "\n", $file_content;
    #print Dumper( @file_lines );
    
    # Extract the individual lines of the licence
    my ( @licence_lines ) = split "\n", $licence_content;
    #print Dumper( @licence_lines );
    
    my $found_first_one = 0;
    my $licence_line = -1;
    my $licence_lines_total = scalar @licence_lines;
    my $file_line = -1;
    my $file_lines_total = scalar @file_lines;
    my $finished = 0;
    foreach my $line ( @file_lines ) {
	$file_line++;
	$_ = $line;
	
	if ( $finished ) {
	    # Keep track of the remaining content
	    $postfix_content .= $line . "\n";
	}
	next if ( $finished || $invalid_licence);
	
	# First check if at least the first line was
	# found. This works in practise only if the first line
	# is distinctive enough.
	if ( $found_first_one == 0 ) {
	    
	    if ( $line =~ m/$licence_content_first_line/ ) {
		$found_first_one = 1;
		$licence_line++;
	    } else {
		# Have not yet found the first line. Keep track of
		# previous content.
		$prefix_content .= $line . "\n";
	    }
	    
	} else {
	    $licence_line++;
	    # Now we are trying to match subsequent lines from
	    # the licence against lines in the file
	    my $licence_content_line = @licence_lines[ $licence_line ];
	    
	    # Stop trying to match lines if we run out of the
	    # licence content
	    if ($licence_line >= $licence_lines_total) {
		$finished = 1;
		$postfix_content .= $line . "\n";
	    } else {
		#print "\nLICENCE CONTENT LINE: $licence_content_line\n";
		#print "\FILE LINE: $line\n";
		
		# Check if this line also matches
		if ( $line  !~ m/$licence_content_line/ ) {
		    $invalid_licence = 1;
		}
	    }
	}
    }


    if ( $found_first_one == 0) {
	return 1, $prefix_content, $postfix_content;
    } elsif ( $invalid_licence == 0 ) {
	return 0, $prefix_content, $postfix_content;
    }
    return 2, $prefix_content, $postfix_content;
}



# Changes the licence of a file, accounting for the type of the
# file. For example, the licence in shell scripts will be prefixed
# with '#'s, in C files with '/*' and postfixed with '*/', etc.
#
# $1 - The content found before the old licence
# $2 - The content found after the old licence
# $3 - The new licence
# $4 - The filename
#
# RETURNS: The new file content
sub change_licence {
    my ( $prefix_file_content, $postfix_file_content, $new_licence, $filename ) = ( @_ );
    my $new_file_content = "";
    my $retval = 1;

    my $Cstyle_prefix = "/*\n";
    my $Cstyle_postfix = "\n*/";

    $Cstyle_prefix = "";
    $Cstyle_postfix = "";

    my $shellstyle_prefix = "# ";
    my $TeXstyle_prefix = "%% ";

    $new_file_content .= $prefix_file_content;

    if ( $filename =~ m/\.c$/) {
	# This is a C file
	$new_file_content .= $Cstyle_prefix;
	$new_file_content .= $new_licence;
	$new_file_content .= $Cstyle_postfix;
    } elsif ( $filename =~ m/\.cpp$/) {
	# This is a C++ file
	$new_file_content .= $Cstyle_prefix;
	$new_file_content .= $new_licence;
	$new_file_content .= $Cstyle_postfix;
    } elsif ( $filename =~ m/\.h$/) {
	# This is a header file
	$new_file_content .= $Cstyle_prefix;
	$new_file_content .= $new_licence;
	$new_file_content .= $Cstyle_postfix;
    } elsif ( $filename =~ m/\.h\.in$/) {
	# This is an autogenerated header file
	$new_file_content .= $Cstyle_prefix;
	$new_file_content .= $new_licence;
	$new_file_content .= $Cstyle_postfix;
    } elsif ( $filename =~ m/\.S$/) {
	# This is an assembly file
	$new_file_content .= $Cstyle_prefix;
	$new_file_content .= $new_licence;
	$new_file_content .= $Cstyle_postfix;
    } elsif ( $filename =~ m/\.sh$/) {
	# This is a shell script
	$new_file_content .= prepend_each_line_with( $new_licence, $shellstyle_prefix);
    } elsif ( $filename =~ m/\.sh\.in$/) {
	# This is an autogenerated shell script
	$new_file_content .= prepend_each_line_with( $new_licence, $shellstyle_prefix);
    } elsif ( $filename =~ m/\.inc$/) {
	# This is a shell library
	$new_file_content .= prepend_each_line_with( $new_licence, $shellstyle_prefix);
    } elsif ( $filename =~ m/\.inc\.in$/) {
	# This is an autogenerated shell library
	$new_file_content .= prepend_each_line_with( $new_licence, $shellstyle_prefix);
    } elsif ( $filename =~ m/\.pl$/) {
	# This is a Perl script
	$new_file_content .= prepend_each_line_with( $new_licence, $shellstyle_prefix);
    } elsif ( $filename =~ m/\.pl\.in$/) {
	# This is an autogenerated Perl script
	$new_file_content .= prepend_each_line_with( $new_licence, $shellstyle_prefix);
    } elsif ( $filename =~ m/\.pm$/) {
	# This is a Perl module
	$new_file_content .= prepend_each_line_with( $new_licence, $shellstyle_prefix);
    } elsif ( $filename =~ m/\.pm\.in$/) {
	# This is an autogenerated Perl module
	$new_file_content .= prepend_each_line_with( $new_licence, $shellstyle_prefix);
    } elsif ( $filename =~ m/Makefile$/) {
	# This is a Makefile
	$new_file_content .= prepend_each_line_with( $new_licence, $shellstyle_prefix);
    } elsif ( $filename =~ m/Makefile\.in$/) {
	# This is an autogenerated Makefile
	$new_file_content .= prepend_each_line_with( $new_licence, $shellstyle_prefix);
    } elsif ( $filename =~ m/\.exp$/) {
	# This is an Expect file
	$new_file_content .= prepend_each_line_with( $new_licence, $shellstyle_prefix);
    } elsif ( $filename =~ m/\.exp\.in$/) {
	# This is an autogenerated Expect file
	$new_file_content .= prepend_each_line_with( $new_licence, $shellstyle_prefix);
    } elsif ( $filename =~ m/\.tex$/) {
	# This is a TeX file
	$new_file_content .= prepend_each_line_with( $new_licence, $TeXstyle_prefix);
    } elsif ( $filename =~ m/\.tex\.in$/) {
	# This is an autogenerated TeX file
	$new_file_content .= prepend_each_line_with( $new_licence, $TeXstyle_prefix);
    } else {
	print "*** Don't know how to handle licence change in filename '$filename'. Ignoring.\n ";
	$retval = 0;
    }

    $new_file_content .= $postfix_file_content;

    return $new_file_content, $retval;
}



sub prepend_each_line_with {
    my ( $content, $prefix ) = ( @_ );
    my @lines = split( "\n", $content );
    my $new_content = "";

    foreach my $line ( @lines ) {
	$new_content .= $prefix . $line . "\n";
    }

    return $new_content;
}



1;
